@@ -7,12 +7,14 @@ use std::iter;
 use std::net::{IpAddr, Ipv4Addr, Ipv6Addr};
 use std::str::FromStr;
 
+use async_hash::Hash;
 use async_trait::async_trait;
 use destream::{de, en};
 use number_general::Number;
 use safecast::{CastFrom, TryCastFrom};
 use serde::de::{Deserialize, Deserializer};
 use serde::ser::{Serialize, Serializer};
+use sha2::digest::{Digest, Output};
 
 use tc_error::*;
 use tcgeneric::{Id, PathLabel, PathSegment, TCPathBuf};
@@ -330,6 +332,12 @@ impl PartialOrd for Link {
     }
 }
 
+impl<'a, D: Digest> Hash<D> for &'a Link {
+    fn hash(self) -> Output<D> {
+        Hash::<D>::hash(self.to_string().as_str())
+    }
+}
+
 impl From<LinkHost> for Link {
     fn from(host: LinkHost) -> Link {
         Link {
@@ -374,6 +382,7 @@ impl TryCastFrom<Value> for Link {
     fn can_cast_from(value: &Value) -> bool {
         match value {
             Value::Bytes(_) => false,
+            Value::Email(email) => Self::from_str(email.get_domain()).is_ok(),
             Value::Id(_) => true,
             Value::Link(_) => true,
             Value::None => true,
@@ -390,6 +399,7 @@ impl TryCastFrom<Value> for Link {
     fn opt_cast_from(value: Value) -> Option<Self> {
         match value {
             Value::Bytes(_) => None,
+            Value::Email(email) => Self::from_str(email.get_domain()).ok(),
             Value::Id(id) => Some(TCPathBuf::from(id).into()),
             Value::Link(l) => Some(l),
             Value::None => Some(TCPathBuf::default().into()),
