@@ -4,18 +4,23 @@ use std::cmp::Ordering;
 use std::convert::{TryFrom, TryInto};
 use std::fmt;
 use std::iter::FromIterator;
+use std::ops::Deref;
 use std::str::FromStr;
 
+use async_hash::Hash;
 use async_trait::async_trait;
 use bytes::Bytes;
 use destream::de;
 use destream::de::Error as DestreamError;
 use destream::de::Visitor as DestreamVisitor;
 use destream::en;
+use email_address_parser::EmailAddress;
 use log::debug;
 use safecast::{CastFrom, CastInto, TryCastFrom, TryCastInto};
 use serde::de::{Deserialize, Deserializer, Error as SerdeError};
 use serde::ser::{Serialize, SerializeMap, Serializer};
+use sha2::digest::generic_array::GenericArray;
+use sha2::digest::{Digest, Output};
 use uuid::Uuid;
 
 use tc_error::*;
@@ -36,6 +41,7 @@ const PREFIX: PathLabel = path_label(&["state", "scalar", "value"]);
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub enum ValueType {
     Bytes,
+    Email,
     Id,
     Link,
     None,
@@ -70,6 +76,16 @@ impl ValueType {
                     .map(Value::Bytes),
                 other => Err(TCError::bad_request("cannot cast into Bytes from", other)),
             },
+            Self::Email => match value {
+                Value::Email(email) => Ok(Value::Email(email)),
+                Value::Id(id) => parse_email(id.as_str())
+                    .map(Value::Email)
+                    .ok_or_else(|| TCError::bad_request("cannot cast into Email from", id)),
+                Value::String(s) => parse_email(&s)
+                    .map(Value::Email)
+                    .ok_or_else(|| TCError::bad_request("cannot cast into Email from", s)),
+                other => Err(TCError::bad_request("cannot cast into Email from", other)),
+            },
             Self::Id => value.try_cast_into(on_err).map(Value::Id),
             Self::Link => value.try_cast_into(on_err).map(Value::String),
             Self::None => Ok(Value::None),
@@ -81,12 +97,14 @@ impl ValueType {
             Self::Tuple => value.try_cast_into(on_err).map(Value::Tuple),
             Self::Value => Ok(value),
             Self::Version => match value {
+                Value::Id(id) => id.as_str().parse().map(Value::Version),
                 Value::String(s) => s.parse().map(Value::Version),
                 Value::Tuple(t) => {
                     let (maj, min, rev) =
                         t.try_cast_into(|t| TCError::bad_request("invalid semantic version", t))?;
                     Ok(Value::Version(Version::from((maj, min, rev))))
                 }
+                Value::Version(v) => Ok(Value::Version(v)),
                 other => Err(TCError::bad_request("cannot cast into Version from", other)),
             },
         }
@@ -117,6 +135,7 @@ impl NativeClass for ValueType {
             } else if path.len() == 4 {
                 match path[3].as_str() {
                     "bytes" => Some(Self::Bytes),
+                    "email" => Some(Self::Email),
                     "id" => Some(Self::Id),
                     "link" => Some(Self::Link),
                     "number" => Some(Self::Number(NT::Number)),
@@ -175,6 +194,7 @@ impl NativeClass for ValueType {
 
         match self {
             Self::Bytes => prefix.append(label("bytes")),
+            Self::Email => prefix.append(label("email")),
             Self::Id => prefix.append(label("id")),
             Self::Link => prefix.append(label("link")),
             Self::None => prefix.append(label("none")),
@@ -259,6 +279,9 @@ impl Ord for ValueType {
             (Self::String, _) => Greater,
             (_, Self::String) => Less,
 
+            (Self::Email, _) => Greater,
+            (_, Self::Email) => Less,
+
             (Self::Id, _) => Greater,
             (_, Self::Id) => Less,
 
@@ -346,6 +369,7 @@ impl fmt::Display for ValueType {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
             Self::Bytes => f.write_str("type Bytes"),
+            Self::Email => f.write_str("type Email"),
             Self::Id => f.write_str("type Id"),
             Self::Link => f.write_str("type Link"),
             Self::None => f.write_str("type None"),
@@ -407,6 +431,7 @@ impl DestreamVisitor for ValueTypeVisitor {
 #[derive(Clone, Eq, PartialEq)]
 pub enum Value {
     Bytes(Bytes),
+    Email(EmailAddress),
     Link(Link),
     Id(Id),
     None,
@@ -459,6 +484,7 @@ impl Value {
 
         match class {
             VT::Bytes => self.opt_cast_into().map(Self::Bytes),
+            VT::Email => self.opt_cast_into().map(Self::Email),
             VT::Id => self.opt_cast_into().map(Self::Id),
             VT::Link => self.opt_cast_into().map(Self::Link),
             VT::None => Some(Self::None),
@@ -489,6 +515,7 @@ impl Instance for Value {
         use ValueType as VT;
         match self {
             Self::Bytes(_) => VT::Bytes,
+            Self::Email(_) => VT::Email,
             Self::Id(_) => VT::Id,
             Self::Link(_) => VT::Link,
             Self::None => VT::None,
@@ -500,6 +527,28 @@ impl Instance for Value {
     }
 }
 
+impl<D: Digest> Hash<D> for Value {
+    fn hash(self) -> Output<D> {
+        Hash::<D>::hash(&self)
+    }
+}
+
+impl<'a, D: Digest> Hash<D> for &'a Value {
+    fn hash(self) -> Output<D> {
+        match self {
+            Value::Bytes(bytes) => D::digest(bytes),
+            Value::Email(email) => Hash::<D>::hash(email.to_string()),
+            Value::Id(id) => Hash::<D>::hash(id),
+            Value::Link(link) => Hash::<D>::hash(link),
+            Value::None => GenericArray::default(),
+            Value::Number(n) => Hash::<D>::hash(*n),
+            Value::String(s) => Hash::<D>::hash(s.as_str()),
+            Value::Tuple(tuple) => Hash::<D>::hash(tuple.deref()),
+            Value::Version(v) => Hash::<D>::hash(*v),
+        }
+    }
+}
+
 impl<'de> Deserialize<'de> for Value {
     fn deserialize<D: Deserializer<'de>>(deserializer: D) -> Result<Self, D::Error> {
         deserializer.deserialize_any(ValueVisitor)
@@ -514,6 +563,11 @@ impl Serialize for Value {
                 map.serialize_entry(&self.class().path().to_string(), &base64::encode(&bytes))?;
                 map.end()
             }
+            Self::Email(email) => {
+                let mut map = serializer.serialize_map(Some(1))?;
+                map.serialize_entry(&self.class().path().to_string(), &email.to_string())?;
+                map.end()
+            }
             Self::Id(id) => {
                 let mut map = serializer.serialize_map(Some(1))?;
                 map.serialize_entry(id.as_str(), &EMPTY_SEQ)?;
@@ -565,6 +619,11 @@ impl<'en> en::ToStream<'en> for Value {
 
         match self {
             Self::Bytes(bytes) => encoder.encode_bytes(bytes),
+            Self::Email(email) => {
+                let mut map = encoder.encode_map(Some(1))?;
+                map.encode_entry(self.class().path().to_string(), email.to_string())?;
+                map.end()
+            }
             Self::Id(id) => {
                 let mut map = encoder.encode_map(Some(1))?;
                 map.encode_entry(id, &EMPTY_SEQ)?;
@@ -597,6 +656,11 @@ impl<'en> en::IntoStream<'en> for Value {
 
         match self {
             Self::Bytes(bytes) => encoder.encode_bytes(&bytes),
+            Self::Email(ref email) => {
+                let mut map = encoder.encode_map(Some(1))?;
+                map.encode_entry(self.class().path().to_string(), email.to_string())?;
+                map.end()
+            }
             Self::Id(id) => {
                 let mut map = encoder.encode_map(Some(1))?;
                 map.encode_entry(id, &EMPTY_SEQ)?;
@@ -742,6 +806,26 @@ impl TryCastFrom<Value> for Bytes {
     }
 }
 
+impl TryCastFrom<Value> for EmailAddress {
+    fn can_cast_from(value: &Value) -> bool {
+        match value {
+            Value::Email(_) => true,
+            Value::Id(id) => parse_email(id.as_str()).is_some(),
+            Value::String(s) => parse_email(s).is_some(),
+            _ => false,
+        }
+    }
+
+    fn opt_cast_from(value: Value) -> Option<Self> {
+        match value {
+            Value::Email(email) => Some(email),
+            Value::Id(id) => parse_email(id.as_str()),
+            Value::String(s) => parse_email(&s),
+            _ => None,
+        }
+    }
+}
+
 impl TryCastFrom<Value> for Id {
     fn can_cast_from(value: &Value) -> bool {
         match value {
@@ -764,6 +848,7 @@ impl TryCastFrom<Value> for Number {
     fn can_cast_from(value: &Value) -> bool {
         match value {
             Value::Bytes(_) => false,
+            Value::Email(_) => false,
             Value::Id(id) => f64::from_str(id.as_str()).is_ok(),
             Value::Link(_) => false,
             Value::None => true,
@@ -781,6 +866,7 @@ impl TryCastFrom<Value> for Number {
     fn opt_cast_from(value: Value) -> Option<Self> {
         match value {
             Value::Bytes(_) => None,
+            Value::Email(_) => None,
             Value::Id(id) => f64::from_str(id.as_str())
                 .map(Float::F64)
                 .map(Self::Float)
@@ -793,8 +879,14 @@ impl TryCastFrom<Value> for Number {
                 let im = Self::opt_cast_from(t.pop().unwrap());
                 let re = Self::opt_cast_from(t.pop().unwrap());
                 match (re, im) {
+                    (Some(Number::Float(Float::F32(re))), Some(Number::Float(Float::F32(im)))) => {
+                        let c = Complex::from([re, im]);
+                        Some(Self::Complex(c))
+                    }
                     (Some(re), Some(im)) => {
-                        let c = Complex::cast_from((re, im));
+                        let re = f64::cast_from(re);
+                        let im = f64::cast_from(im);
+                        let c = Complex::from([re, im]);
                         Some(Self::Complex(c))
                     }
                     _ => None,
@@ -877,65 +969,39 @@ impl TryCastFrom<Value> for Version {
     }
 }
 
-impl TryCastFrom<Value> for bool {
-    fn can_cast_from(value: &Value) -> bool {
-        Number::can_cast_from(value)
-    }
+macro_rules! cast_real {
+    ($n:ty) => {
+        impl TryCastFrom<Value> for $n {
+            fn can_cast_from(value: &Value) -> bool {
+                match value {
+                    Value::Number(_) => true,
+                    Value::String(s) => Self::from_str(s).is_ok(),
+                    _ => false,
+                }
+            }
 
-    fn opt_cast_from(value: Value) -> Option<Self> {
-        Number::opt_cast_from(value).map(|n| n.cast_into())
-    }
+            fn opt_cast_from(value: Value) -> Option<Self> {
+                match value {
+                    Value::Number(n) => n.opt_cast_into(),
+                    Value::String(s) => Self::from_str(&s).ok(),
+                    _ => None,
+                }
+            }
+        }
+    };
 }
 
-impl TryCastFrom<Value> for u8 {
-    fn can_cast_from(value: &Value) -> bool {
-        Number::can_cast_from(value)
-    }
-
-    fn opt_cast_from(value: Value) -> Option<Self> {
-        Number::opt_cast_from(value).map(|n| n.cast_into())
-    }
-}
-
-impl TryCastFrom<Value> for usize {
-    fn can_cast_from(value: &Value) -> bool {
-        Number::can_cast_from(value)
-    }
-
-    fn opt_cast_from(value: Value) -> Option<Self> {
-        Number::opt_cast_from(value).map(|n| n.cast_into())
-    }
-}
-
-impl TryCastFrom<Value> for u32 {
-    fn can_cast_from(value: &Value) -> bool {
-        Number::can_cast_from(value)
-    }
-
-    fn opt_cast_from(value: Value) -> Option<Self> {
-        Number::opt_cast_from(value).map(|n| n.cast_into())
-    }
-}
-
-impl TryCastFrom<Value> for u64 {
-    fn can_cast_from(value: &Value) -> bool {
-        Number::can_cast_from(value)
-    }
-
-    fn opt_cast_from(value: Value) -> Option<Self> {
-        Number::opt_cast_from(value).map(|n| n.cast_into())
-    }
-}
-
-impl TryCastFrom<Value> for i64 {
-    fn can_cast_from(value: &Value) -> bool {
-        Number::can_cast_from(value)
-    }
-
-    fn opt_cast_from(value: Value) -> Option<Self> {
-        Number::opt_cast_from(value).map(|n| n.cast_into())
-    }
-}
+cast_real!(bool);
+cast_real!(f32);
+cast_real!(f64);
+cast_real!(u8);
+cast_real!(u16);
+cast_real!(u32);
+cast_real!(u64);
+cast_real!(usize);
+cast_real!(i16);
+cast_real!(i32);
+cast_real!(i64);
 
 impl TryCastFrom<Value> for TCPathBuf {
     fn can_cast_from(value: &Value) -> bool {
@@ -1048,6 +1114,7 @@ impl fmt::Debug for Value {
         match self {
             Self::Bytes(bytes) => write!(f, "({} bytes)", bytes.len()),
             Self::Id(id) => write!(f, "{}: {:?}", ValueType::Id, id.as_str()),
+            Self::Email(email) => write!(f, "{}: {:?}", ValueType::Email, email),
             Self::Link(link) => write!(f, "{}: {:?}", ValueType::Link, link),
             Self::None => f.write_str("None"),
             Self::Number(n) => fmt::Debug::fmt(n, f),
@@ -1070,6 +1137,7 @@ impl fmt::Display for Value {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
             Self::Bytes(bytes) => write!(f, "({} bytes)", bytes.len()),
+            Self::Email(email) => fmt::Display::fmt(email, f),
             Self::Id(id) => f.write_str(id.as_str()),
             Self::Link(link) => fmt::Display::fmt(link, f),
             Self::None => f.write_str("None"),
@@ -1115,6 +1183,15 @@ impl ValueVisitor {
                     .map(Value::Bytes)
                     .map_err(serde::de::Error::custom)
             }
+            VT::Email => {
+                let email: &str = map.next_value()?;
+                parse_email(email).map(Value::Email).ok_or_else(|| {
+                    serde::de::Error::custom(format!(
+                        "invalid value: {}, expected an email address",
+                        email
+                    ))
+                })
+            }
             VT::Id => {
                 let id: &str = map.next_value()?;
                 Id::from_str(id).map(Value::Id).map_err(A::Error::custom)
@@ -1169,6 +1246,12 @@ impl ValueVisitor {
                 let bytes = map.next_value(()).await?;
                 Ok(Value::Bytes(bytes))
             }
+            VT::Email => {
+                let email: String = map.next_value(()).await?;
+                parse_email(&email)
+                    .map(Value::Email)
+                    .ok_or_else(|| de::Error::invalid_value(email, "an email address"))
+            }
             VT::Id => {
                 let id = map.next_value(()).await?;
                 Ok(Value::Id(id))
@@ -1439,3 +1522,7 @@ impl destream::de::Visitor for ValueVisitor {
         Ok(Value::Tuple(value.into()))
     }
 }
+
+fn parse_email(s: &str) -> Option<EmailAddress> {
+    EmailAddress::parse(s.as_ref(), None)
+}
